package cz.muni.fi.pv112.cv5;

import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.imageio.ImageIO;
import org.joml.Matrix3f;
import org.joml.Matrix4f;
import org.joml.Vector3f;
import org.lwjgl.BufferUtils;
import org.lwjgl.Version;
import org.lwjgl.glfw.GLFWErrorCallback;
import org.lwjgl.glfw.GLFWVidMode;
import org.lwjgl.opengl.GL;
import org.lwjgl.system.MemoryStack;
import static org.lwjgl.glfw.Callbacks.glfwFreeCallbacks;
import static org.lwjgl.glfw.GLFW.*;
import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL12.*;
import static org.lwjgl.opengl.GL13.*;
import static org.lwjgl.opengl.GL14.*;
import static org.lwjgl.opengl.GL15.*;
import static org.lwjgl.opengl.GL20.*;
import static org.lwjgl.opengl.GL30.*;
import static org.lwjgl.system.MemoryStack.stackPush;
import static org.lwjgl.system.MemoryUtil.NULL;

public class Cv5 {

    private static final int SIZEOF_AXES_VERTEX = 6 * Float.BYTES;
    private static final int COLOR_OFFSET = 3 * Float.BYTES;

    private static final float AXES[] = {
        // .. position .......... color ....
        // x axis
        1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
        0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
        // y axis
        0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f,
        0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f,
        // z axis
        0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f,
        0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f
    };

    private static final int SIZEOF_MODEL_VERTEX = 8 * Float.BYTES;
    private static final int NORMAL_OFFSET = 3 * Float.BYTES;
    private static final int TEXCOORD_OFFSET = 6 * Float.BYTES;

    private Camera camera;

    // the window handle
    private long window;

    // window size
    private int width;
    private int height;
    private boolean resized = false;

    // animation
    private boolean animate = false;
    private float t = 0f;

    // rendering mode
    private int mode = GL_FILL;

    // model
    private ObjLoader cube;
    private ObjLoader teapot;

    // our OpenGL resources
    private int axesBuffer;
    private int cubeBuffer;
    private int teapotBuffer;
    private int axesArray;
    private int cubeArray;
    private int teapotArray;

    // our GLSL resources
    private int axesProgram;
    private int axesAspectUniformLoc;
    private int axesLengthUniformLoc;
    private int axesMvpUniformLoc;

    private int modelProgram;
    private int modelMvpLoc;
    private int modelNLoc;
    private int modelModelLoc;

    private int lightPositionLoc;
    private int lightAmbientColorLoc;
    private int lightDiffuseColorLoc;
    private int lightSpecularColorLoc;

    private int materialAmbientColorLoc;
    private int materialDiffuseColorLoc;
    private int materialSpecularColorLoc;
    private int materialShininessLoc;

    private int eyePositionLoc;

    // Fullscreen quad
    // Task 10: make sure you understand why we need them
    private static final float[] SCREEN_QUAD_DATA = {
        // position (2D)
        -1.0f,  1.0f,
        -1.0f, -1.0f,
         1.0f,  1.0f,
         1.0f, -1.0f
    };
    int screenQuadBuffer;
    int screenQuadArray;

    // Framebuffer object
    int fbo_width, fbo_height;
    int fboColorTexture;
    int fboDepthStencilTexture;
    int fbo;

    // Postprocess shader
    int postprocess_program;
    int sceneTex_loc;
    int animation_time_loc;
    
    public static void main(String[] args) {
        new Cv5().run();
    }

    public void run() {
        System.out.println("Hello LWJGL " + Version.getVersion() + "!");

        camera = new Camera();

        initGLFW();
        loop();

        // Free the window callbacks and destroy the window
        glfwFreeCallbacks(window);
        glfwDestroyWindow(window);

        // Terminate GLFW and free the error callback
        glfwTerminate();
        glfwSetErrorCallback(null).free();
    }

    private void initGLFW() {
        // Setup an error callback. The default implementation
        // will print the error message in System.err.
        GLFWErrorCallback.createPrint(System.err).set();

        // Initialize GLFW. Most GLFW functions will not work before doing this.
        if (!glfwInit()) {
            throw new IllegalStateException("Unable to initialize GLFW");
        }

        // Configure GLFW
        glfwDefaultWindowHints(); // optional, the current window hints are already the default
        glfwWindowHint(GLFW_VISIBLE, GLFW_FALSE); // the window will stay hidden after creation
        glfwWindowHint(GLFW_RESIZABLE, GLFW_TRUE); // the window will be resizable
        glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3); // request the OpenGL 3.3 core profile context
        glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
        glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

        // set initial width and height
        width = 640;
        height = 480;

        // Create the window
        window = glfwCreateWindow(width, height, "Hello World!", NULL, NULL);
        if (window == NULL) {
            throw new RuntimeException("Failed to create the GLFW window");
        }

        // Setup a key callback. It will be called every time a key is pressed, repeated or released.
        glfwSetKeyCallback(window, (window, key, scancode, action, mods) -> {
            if (action == GLFW_RELEASE) {
                switch (key) {
                    case GLFW_KEY_ESCAPE:
                        glfwSetWindowShouldClose(window, true); // We will detect this in the rendering loop
                        break;
                    case GLFW_KEY_A:
                        animate = !animate;
                        break;
                    case GLFW_KEY_T:
                        // TODO toggle fullscreen
                        break;
                    case GLFW_KEY_L:
                        mode = GL_LINE;
                        break;
                    case GLFW_KEY_F:
                        mode = GL_FILL;
                        break;
                }
            }
        });

        glfwSetMouseButtonCallback(window, (window, button, action, mods) -> {
            if (action == GLFW_PRESS) {
                if (button == GLFW_MOUSE_BUTTON_1) {
                    camera.updateMouseButton(Camera.Button.LEFT, true);
                } else if (button == GLFW_MOUSE_BUTTON_2) {
                    camera.updateMouseButton(Camera.Button.RIGHT, true);
                }
            } else if (action == GLFW_RELEASE) {
                if (button == GLFW_MOUSE_BUTTON_1) {
                    camera.updateMouseButton(Camera.Button.LEFT, false);
                } else if (button == GLFW_MOUSE_BUTTON_2) {
                    camera.updateMouseButton(Camera.Button.RIGHT, false);
                }
            }
        });

        glfwSetCursorPosCallback(window, (window, xpos, ypos) -> {
            camera.updateMousePosition(xpos, ypos);
        });

        // add window size callback
        glfwSetWindowSizeCallback(window, (window, width, height) -> {
            this.width = width;
            this.height = height;
            resized = true;
        });

        // Get the thread stack and push a new frame
        try (MemoryStack stack = stackPush()) {
            IntBuffer pWidth = stack.mallocInt(1); // int*
            IntBuffer pHeight = stack.mallocInt(1); // int*

            // Get the window size passed to glfwCreateWindow
            glfwGetWindowSize(window, pWidth, pHeight);

            // Get the resolution of the primary monitor
            GLFWVidMode vidmode = glfwGetVideoMode(glfwGetPrimaryMonitor());

            // Center the window
            glfwSetWindowPos(
                    window,
                    (vidmode.width() - pWidth.get(0)) / 2,
                    (vidmode.height() - pHeight.get(0)) / 2
            );
        } // the stack frame is popped automatically

        // Make the OpenGL context current
        glfwMakeContextCurrent(window);
        // Enable v-sync
        glfwSwapInterval(1);

        // Make the window visible
        glfwShowWindow(window);

        // This line is critical for LWJGL's interoperation with GLFW's
        // OpenGL context, or any context that is managed externally.
        // LWJGL detects the context that is current in the current thread,
        // creates the GLCapabilities instance and makes the OpenGL
        // bindings available for use.
        GL.createCapabilities();
    }

    private void loop() {
        // Prepare data for rendering
        init();
        
        initPostprocessing();

        // Run the rendering loop until the user has attempted to close
        // the window or has pressed the ESCAPE key.
        while (!glfwWindowShouldClose(window)) {
            render();

            glfwSwapBuffers(window); // swap the color buffers

            // Poll for window events. The key callback above will only be
            // invoked during this call.
            glfwPollEvents();
        }
    }

    private void init() {
        // empty scene color
        glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        glLineWidth(3.0f); // makes lines thicker

        glEnable(GL_DEPTH_TEST);

        // load GLSL program (vertex, fragment shaders) and textures
        try {
            axesProgram = loadProgram("/resources/shaders/axes.vs.glsl",
                    "/resources/shaders/axes.fs.glsl");
            modelProgram = loadProgram("/resources/shaders/model.vs.glsl",
                    "/resources/shaders/model.fs.glsl");
        } catch (IOException ex) {
            Logger.getLogger(Cv5.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        }

        // get uniform locations
        // axes program uniforms
        axesAspectUniformLoc = glGetUniformLocation(axesProgram, "aspect");
        axesLengthUniformLoc = glGetUniformLocation(axesProgram, "len");
        axesMvpUniformLoc = glGetUniformLocation(axesProgram, "MVP");

        // model program uniforms
        modelMvpLoc = glGetUniformLocation(modelProgram, "MVP");
        modelNLoc = glGetUniformLocation(modelProgram, "N");
        modelModelLoc = glGetUniformLocation(modelProgram, "model");

        lightPositionLoc = glGetUniformLocation(modelProgram, "lightPosition");
        lightAmbientColorLoc = glGetUniformLocation(modelProgram, "lightAmbientColor");
        lightDiffuseColorLoc = glGetUniformLocation(modelProgram, "lightDiffuseColor");
        lightSpecularColorLoc = glGetUniformLocation(modelProgram, "lightSpecularColor");

        eyePositionLoc = glGetUniformLocation(modelProgram, "eyePosition");

        materialAmbientColorLoc = glGetUniformLocation(modelProgram, "materialAmbientColor");
        materialDiffuseColorLoc = glGetUniformLocation(modelProgram, "materialDiffuseColor");
        materialSpecularColorLoc = glGetUniformLocation(modelProgram, "materialSpecularColor");
        materialShininessLoc = glGetUniformLocation(modelProgram, "materialShininess");

        // create buffers with geometry
        int[] buffers = new int[3];
        glGenBuffers(buffers);
        axesBuffer = buffers[0];
        cubeBuffer = buffers[1];
        teapotBuffer = buffers[2];

        // fill a buffers with geometry
        glBindBuffer(GL_ARRAY_BUFFER, axesBuffer);
        glBufferData(GL_ARRAY_BUFFER, AXES, GL_STATIC_DRAW);

        // load cube and fill buffer with cube data
        cube = new ObjLoader("/resources/models/cube.obj");
        try {
            cube.load();
        } catch (IOException ex) {
            Logger.getLogger(Cv5.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        }
        int cubeBufferLength = 3 * 8 * cube.getTriangleCount();
        FloatBuffer cubeData = BufferUtils.createFloatBuffer(cubeBufferLength);
        for (int f = 0; f < cube.getTriangleCount(); f++) {
            int[] pi = cube.getVertexIndices().get(f);
            int[] ni = cube.getNormalIndices().get(f);
            int[] ti = cube.getTexcoordIndices().get(f);
            for (int i = 0; i < 3; i++) {
                float[] position = cube.getVertices().get(pi[i]);
                float[] normal = cube.getNormals().get(ni[i]);
                float[] texcoord = cube.getTexcoords().get(ti[i]);
                cubeData.put(position);
                cubeData.put(normal);
                cubeData.put(texcoord);
            }
        }
        cubeData.rewind();
        glBindBuffer(GL_ARRAY_BUFFER, cubeBuffer);
        glBufferData(GL_ARRAY_BUFFER, cubeData, GL_STATIC_DRAW);

        // load teapot and fill buffer with teapot data
        teapot = new ObjLoader("/resources/models/teapot.obj");
        try {
            teapot.load();
        } catch (IOException ex) {
            Logger.getLogger(Cv5.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        }
        int teapotLength = 3 * 8 * teapot.getTriangleCount();
        FloatBuffer teapotData = BufferUtils.createFloatBuffer(teapotLength);
        for (int f = 0; f < teapot.getTriangleCount(); f++) {
            int[] pi = teapot.getVertexIndices().get(f);
            int[] ni = teapot.getNormalIndices().get(f);
            int[] ti = teapot.getTexcoordIndices().get(f);
            for (int i = 0; i < 3; i++) {
                float[] position = teapot.getVertices().get(pi[i]);
                float[] normal = teapot.getNormals().get(ni[i]);
                float[] texcoord = teapot.getTexcoords().get(ti[i]);
                teapotData.put(position);
                teapotData.put(normal);
                teapotData.put(texcoord);
            }
        }
        teapotData.rewind();
        glBindBuffer(GL_ARRAY_BUFFER, teapotBuffer);
        glBufferData(GL_ARRAY_BUFFER, teapotData, GL_STATIC_DRAW);

        // clear buffer binding, so that other code doesn't presume it (easier error detection)
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        // create a vertex array object for the geometry
        int[] arrays = new int[3];
        glGenVertexArrays(arrays);
        axesArray = arrays[0];
        cubeArray = arrays[1];
        teapotArray = arrays[2];

        // get axes program attributes
        int positionAttribLoc = glGetAttribLocation(axesProgram, "position");
        int colorAttribLoc = glGetAttribLocation(axesProgram, "color");
        // bind axes buffer
        glBindVertexArray(axesArray);
        glBindBuffer(GL_ARRAY_BUFFER, axesBuffer);
        // stride and offset are employed as both position and color data reside in the same vertex buffer
        glEnableVertexAttribArray(positionAttribLoc);
        glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, false, SIZEOF_AXES_VERTEX, 0);
        glEnableVertexAttribArray(colorAttribLoc);
        glVertexAttribPointer(colorAttribLoc, 3, GL_FLOAT, false, SIZEOF_AXES_VERTEX, COLOR_OFFSET);

        // get cube program attributes
        positionAttribLoc = glGetAttribLocation(modelProgram, "position");
        int normalAttribLoc = glGetAttribLocation(modelProgram, "normal");
        int texcoordAttribLoc = glGetAttribLocation(modelProgram, "texcoord");

        // bind cube buffer
        glBindVertexArray(cubeArray);
        glBindBuffer(GL_ARRAY_BUFFER, cubeBuffer);
        glEnableVertexAttribArray(positionAttribLoc);
        glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, 0);
        glEnableVertexAttribArray(normalAttribLoc);
        glVertexAttribPointer(normalAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, NORMAL_OFFSET);
        glEnableVertexAttribArray(texcoordAttribLoc);
        glVertexAttribPointer(texcoordAttribLoc, 2, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, TEXCOORD_OFFSET);

        // bind teapot buffer
        glBindVertexArray(teapotArray);
        glBindBuffer(GL_ARRAY_BUFFER, teapotBuffer);
        glEnableVertexAttribArray(positionAttribLoc);
        glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, 0);
        glEnableVertexAttribArray(normalAttribLoc);
        glVertexAttribPointer(normalAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, NORMAL_OFFSET);
        glEnableVertexAttribArray(texcoordAttribLoc);
        glVertexAttribPointer(texcoordAttribLoc, 2, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, TEXCOORD_OFFSET);

        // clear bindings, so that other code doesn't presume it (easier error detection)
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        glBindVertexArray(0);
    }
    
    private void initPostprocessing() {
        // Create post-processing shader program
        // Task 10: examine initialization of all objects needed for off-screen rendering and post-processing
        try {
            postprocess_program = loadProgram("/resources/shaders/postprocess.vs.glsl", "/resources/shaders/postprocess.fs.glsl");
        } catch (IOException ex) {
            Logger.getLogger(Cv5.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        }

        // Get attribute and uniform locations
        int position_loc = glGetAttribLocation(postprocess_program, "position");
        sceneTex_loc = glGetUniformLocation(postprocess_program, "sceneTex");
        animation_time_loc = glGetUniformLocation(postprocess_program, "animation_time");

        // Set the sceneTex texture unit to 0
        glUseProgram(postprocess_program);
        glUniform1i(sceneTex_loc, 0);
        glUseProgram(0);

        // Prepare screen quad geometry and VAO
        screenQuadBuffer = glGenBuffers();
        glBindBuffer(GL_ARRAY_BUFFER, screenQuadBuffer);
        glBufferData(GL_ARRAY_BUFFER, SCREEN_QUAD_DATA, GL_STATIC_DRAW);

        screenQuadArray = glGenVertexArrays();
        glBindVertexArray(screenQuadArray);
        glEnableVertexAttribArray(position_loc);
        glVertexAttribPointer(position_loc, 2, GL_FLOAT, false, 0, 0);
        glBindVertexArray(0);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        // Prepare FBO textures
        fboColorTexture = glGenTextures();
        fboDepthStencilTexture = glGenTextures();
        resize_fbo_textures();

        // Set parameters of the color texture. We do not read the depth texture for rendering,
        // so we do not set its parameters.
        glBindTexture(GL_TEXTURE_2D, fboColorTexture);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glBindTexture(GL_TEXTURE_2D, 0);

        // Prepare FBO
        fbo = glGenFramebuffers();
        glBindFramebuffer(GL_FRAMEBUFFER, fbo);
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, fboColorTexture, 0);
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT, GL_TEXTURE_2D, fboDepthStencilTexture, 0);
        glDrawBuffers(GL_COLOR_ATTACHMENT0);
        int status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
        if (status != GL_FRAMEBUFFER_COMPLETE) {
            Logger.getLogger(Cv5.class.getName()).log(Level.SEVERE, "Failed to create a complete framebuffer");
            System.exit(1);
        }
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
    }
    
    private void render() {
        // Task 13: fix this method, so that only the original scene is rendered as a wireframe (won't tell you where ☺)
        // Resize OpenGL viewport, i.e., the (bitmap) extents to that is the
        // OpenGL screen space [-1, 1] mapped.
        if (resized) {
            // Task 10: replace glViewport(...) with resize_fbo_textures()
            glViewport(0, 0, width, height);
            resized = false;
        }

        // animate variables
        if (animate) {
            t += 0.02f;
        }

        // Task 10: set rendering to off-screen FBO with glBindFramebuffer(GL_FRAMEBUFFER, fbo)
        //          set proper viewport size for rendering to the FBO with glViewport(0, 0, <width>, <height>)
        //              see method resize_fbo_textures() regarding <width> and <height> 
        glPolygonMode(GL_FRONT_AND_BACK, mode);

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Task 10: set proper width/height ratio for projection, consider FBO size
        Matrix4f projection = new Matrix4f()
                .perspective((float) Math.toRadians(60f), width / (float) height, 1, 500);

        Matrix4f view = new Matrix4f()
                .lookAt(camera.getEyePosition(), new Vector3f(0, 0, 0), new Vector3f(0, 1, 0));

        Material mat = new Material(new Vector3f(1, 1, 1), new Vector3f(1, 1, 1), new Vector3f(1, 1, 1), 45);

        // Task 9: use teapot instead of a cube
        drawModel(new Matrix4f(), view, projection, cubeArray, cube.getTriangleCount() * 3, mat);

        // Task 10: setup rendering to window (default/"main") framebuffer:
        //              bind window framebuffer as a current target with glBindframeBuffer(GL_FRAMEBUFFER, 0)
        //              set proper viewport size for rendering to window
        //              clear color information in window framebuffer (we don't care about depth while drawing the fullscreen quad)
        //              also disable depth test with glDisable(GL_DEPTH_TEST)
        
        // Task 10: render one big rectangle over the entire screen:
        //              use postprocess_program to use correct shaders with glUseProgram(...)
        //              activate GL_TEXTURE0 texturing unit
        //              bind fboColorTexture texture into it (we have previously set the sampler in postprocess_program to use 0)
        //              bind screenQuadArray as current VertexArray with glBindVertexArray(...)
        //              draw 4 points from the vertex array as a GL_TRIANGLE_STRIP
        // Task 12: set uniform variable "animation_time" to our variable t (think WHEN to assign it, the order of commands matters)
        
        // Task 10: examine clean up:
        //              unbind VertexArray so it's easier to detect errors by calling glBindVertexArray(0)
        //              unbind shader program so it's easier to detect errors by calling glUseProgram(0)
        //              re-enable depth test for subsequent rendering
        glBindVertexArray(0);
        glUseProgram(0);
        glEnable(GL_DEPTH_TEST);
    }

    // Task 10: examine
    private void resize_fbo_textures() {
        // Task 10: later, you can try rendering in different resolutions by changing these
	fbo_width = width;
	fbo_height = height;
	glBindTexture(GL_TEXTURE_2D, fboColorTexture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, fbo_width, fbo_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, (float[])null);
	glBindTexture(GL_TEXTURE_2D, fboDepthStencilTexture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH24_STENCIL8, fbo_width, fbo_height, 0, GL_DEPTH_STENCIL, GL_UNSIGNED_INT_24_8, (float[])null);
	glBindTexture(GL_TEXTURE_2D, 0);
    }
    
    private void drawModel(Matrix4f model, Matrix4f view, Matrix4f projection, int vao, int count, Material material) {
        drawModel(model, view, projection, vao, 0, count, material);
    }

    private void drawModel(Matrix4f model, Matrix4f view, Matrix4f projection, int vao, int offset, int count, Material material) {
        // compute model-view-projection matrix
        Matrix4f mvp = new Matrix4f(projection)
                .mul(view)
                .mul(model);

        // compute normal matrix
        Matrix3f n = model.get3x3(new Matrix3f())
                .invert()
                .transpose();

        glUseProgram(modelProgram);
        glBindVertexArray(vao); // bind vertex array to draw

        glUniform4f(lightPositionLoc, 2, 5, 3, 1);
        glUniform3f(lightAmbientColorLoc, 0.3f, 0.3f, 0.3f);
        glUniform3f(lightDiffuseColorLoc, 1, 1, 1);
        glUniform3f(lightSpecularColorLoc, 1, 1, 1);

        glUniform3f(eyePositionLoc, camera.getEyePosition().x, camera.getEyePosition().y, camera.getEyePosition().z);

        if (material != null) {
            glUniform3f(materialAmbientColorLoc, material.getAmbientColor().x, material.getAmbientColor().y, material.getAmbientColor().z);
            glUniform3f(materialDiffuseColorLoc, material.getDiffuseColor().x, material.getDiffuseColor().y, material.getDiffuseColor().z);
            glUniform3f(materialSpecularColorLoc, material.getSpecularColor().x, material.getSpecularColor().y, material.getSpecularColor().z);
            glUniform1f(materialShininessLoc, material.getShininess());
        }

        FloatBuffer mvpData = BufferUtils.createFloatBuffer(16);
        FloatBuffer nData = BufferUtils.createFloatBuffer(9);
        FloatBuffer modelData = BufferUtils.createFloatBuffer(16);
        mvp.get(mvpData);
        n.get(nData);
        model.get(modelData);
        glUniformMatrix4fv(modelMvpLoc, false, mvpData); // pass MVP matrix to shader
        glUniformMatrix3fv(modelNLoc, false, nData); // pass Normal matrix to shader
        glUniformMatrix4fv(modelModelLoc, false, modelData); // pass model matrix to shader

        glDrawArrays(GL_TRIANGLES, offset, count);

        glBindTexture(GL_TEXTURE_2D, 0);
        glBindVertexArray(0);
        glUseProgram(0);
    }

    private void drawAxes(Matrix4f modelViewProjection, float length) {
        glUseProgram(axesProgram);
        glBindVertexArray(axesArray);

        FloatBuffer mvpData = BufferUtils.createFloatBuffer(16);
        modelViewProjection.get(mvpData);
        glUniform1f(axesLengthUniformLoc, length);
        glUniform1f(axesAspectUniformLoc, width / (float) height);
        glUniformMatrix4fv(axesMvpUniformLoc, false, mvpData);

        glDrawArrays(GL_LINES, 0, 6);

        glBindVertexArray(0);
        glUseProgram(0);
    }

    private int loadTexture(String filename) throws IOException {
        BufferedImage image = ImageIO.read(Cv5.class.getResourceAsStream(filename));
        byte[] pixels = ((DataBufferByte) image.getRaster().getDataBuffer()).getData();

        int internalFormat;
        int format;
        switch (image.getType()) {
            case BufferedImage.TYPE_3BYTE_BGR:
                internalFormat = GL_RGB;
                format = GL_BGR;
                break;
            case BufferedImage.TYPE_4BYTE_ABGR:
                internalFormat = GL_RGBA;
                format = GL_BGRA;
                pixels = toBGRA(pixels);
                break;
            default:
                throw new IOException("Unknown image type: " + image.getType());
        }

        ByteBuffer textureData = BufferUtils.createByteBuffer(pixels.length);
        textureData.put(pixels);
        textureData.rewind();

        int texture = glGenTextures();
        glBindTexture(GL_TEXTURE_2D, texture);
        glTexImage2D(GL_TEXTURE_2D, 0, internalFormat, image.getWidth(), image.getHeight(), 0, format, GL_UNSIGNED_BYTE, textureData);
        glGenerateMipmap(GL_TEXTURE_2D);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

        // unbind texture
        glBindTexture(GL_TEXTURE_2D, 0);

        return texture;
    }

    private byte[] toBGRA(byte[] abgr) {
        byte[] bgra = new byte[abgr.length];
        for (int i = 0; i < abgr.length; i += 4) {
            bgra[i] = abgr[i + 1];
            bgra[i + 1] = abgr[i + 2];
            bgra[i + 2] = abgr[i + 3];
            bgra[i + 3] = abgr[i];
        }
        return bgra;
    }

    private int loadShader(String filename, int shaderType) throws IOException {
        String source = readAllFromResource(filename);
        int shader = glCreateShader(shaderType);

        // create and compile GLSL shader
        glShaderSource(shader, source);
        glCompileShader(shader);

        // check GLSL shader compile status
        int status = glGetShaderi(shader, GL_COMPILE_STATUS);
        if (status == GL_FALSE) {
            String log = glGetShaderInfoLog(shader);
            System.err.println(log);
        }

        return shader;
    }

    private int loadProgram(String vertexShaderFile, String fragmentShaderFile) throws IOException {
        // load vertex and fragment shaders (GLSL)
        int vs = loadShader(vertexShaderFile, GL_VERTEX_SHADER);
        int fs = loadShader(fragmentShaderFile, GL_FRAGMENT_SHADER);

        // create GLSL program, attach shaders and compile it
        int program = glCreateProgram();
        glAttachShader(program, vs);
        glAttachShader(program, fs);
        glLinkProgram(program);

        int status = glGetProgrami(program, GL_LINK_STATUS);
        if (status == GL_FALSE) {
            String log = glGetProgramInfoLog(program);
            System.err.println(log);
        }

        return program;
    }

    private String readAllFromResource(String resource) throws IOException {
        InputStream is = Cv5.class.getResourceAsStream(resource);
        if (is == null) {
            throw new IOException("Resource not found: " + resource);
        }

        BufferedReader reader = new BufferedReader(new InputStreamReader(is));
        StringBuilder sb = new StringBuilder();

        int c;
        while ((c = reader.read()) != -1) {
            sb.append((char) c);
        }

        return sb.toString();
    }

}
